#include <systemd/sd-bus.h>

#include <security/pam_ext.h>
#include <security/pam_appl.h>
#include <security/pam_modules.h>

#include <json-c/json.h>

#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>
#include <syslog.h>
#include <errno.h>
#include <unistd.h>
#include <pthread.h>
#include <locale.h>
#include <libintl.h>
#include <signal.h>
#include <termios.h>
#include <pwd.h>
#include <limits.h>

#define DBUS_SERVICE "com.deepin.daemon.Authenticate"
#define DBUS_PATH "/com/deepin/daemon/Authenticate"
#define DBUS_INTERFACE "com.deepin.daemon.Authenticate"
#define FINGER_DBUS_PATH "/com/deepin/daemon/Authenticate/Fingerprint"
#define FINGER_DBUS_INTERFACE "com.deepin.daemon.Authenticate.Fingerprint"

#define AUTH_SIGNAL_STATUS "Status"
#define FINGER_SIGNAL_VERIFYSTATUS "VerifyStatus"
#define MAX_BUF_SIZE 256
#define LIMITS_BUF_SIZE 1024

#define UNUSED_VALUE(a) ((void)(a))

#define AUTH_FLAG_MIN (AT_Password)
#define AUTH_FLAG_MAX (AT_Password | AT_Fingerprint | AT_Face | AT_ActiveDirectory)

#define D_DEBUG(pamh, text, ...) \
    if (g_beDebug) { \
        pam_syslog(pamh, LOG_DEBUG, text, ##__VA_ARGS__); \
    }

enum AuthenticateType {
    AT_Password = 0x0001,
    AT_Fingerprint = 0x0002,
    AT_Face = 0x0004,
    AT_ActiveDirectory = 0x00008
};

enum AuthStutaCode{
    ASC_Success = 0, // 认证成功
    ASC_Failure, // 认证失败
    ASC_Cancel, // 认证取消
    ASC_Timeout,        // 认证超时
    ASC_Error,         // 认证错误
    ASC_Verify
};

enum VarifyCode{
    VC_Match = 0, // 认证成功
    VC_NoMatch, // 认证失败
    VC_Error, // 认证取消
    VC_Retry, //重试
    VC_Disconnect //失去连接
};

struct UserData {
    pam_handle_t *pamh;
    sd_bus *bus;
    char id[MAX_BUF_SIZE];
    char username[MAX_BUF_SIZE];
    pthread_t pid;
    bool needCancel;
};

struct Limit {
    int maxTries;
    int numFailures;
    bool locked;
    char unlockTime[MAX_BUF_SIZE];
};

static struct Limit limits[4] = {0};
static const int PASSWORD_LIMIT_INDEX = 0;
static const int FINGERPRINT_LIMIT_INDEX = 1;
static const int FACE_LIMIT_INDEX = 2;
static const int ACTIVE_DIRECTORY_LIMIT_INDEX = 3;

static char *auth_tok = NULL;
pthread_mutex_t userDataMut = PTHREAD_MUTEX_INITIALIZER;

static struct termios *term = NULL;
static struct UserData *userData = NULL;

static char *const TIMEOUT_FLAG = "timeout";
static char *const USER_LOCALE_FLAG = "user_locale";
static char *const DEBUG_FLAG = "debug";

static int g_res = -1;
static int g_beDebug = 0;

static int dbus_method_authenticate(pam_handle_t *pamh, sd_bus *bus, const char *username,
                                    int flags, int timeout, char *id);
static int dbus_method_get_limits(pam_handle_t *pamh, sd_bus *bus, const char *username, char *limits);
static int dbus_method_cancelAuthenticate(pam_handle_t *pamh, sd_bus *bus, const char *id);
static int dbus_method_setPassword(pam_handle_t *pamh, sd_bus *bus, const char *id, const char *password);
static int dbus_method_oneKeyLogin(pam_handle_t *pamh, sd_bus *bus, const char *username, char *id);
static struct pam_response *send_msg(pam_handle_t *pamh, const char *msg, int style);

void pam_clean_func(void *user_data) {
    pthread_mutex_lock(&userDataMut);
    struct UserData *data = *(struct UserData **)user_data;
    *(struct UserData **)user_data = NULL;
    if (!data) {
        pthread_mutex_unlock(&userDataMut);
        return;
    }
    pthread_mutex_unlock(&userDataMut);

    int ret = 0;
    if (data->bus) {
        if (data->needCancel) {
            D_DEBUG(data->pamh, "close authenticate bus!");
            dbus_method_cancelAuthenticate(data->pamh, data->bus, data->id);
        }
        sd_bus_flush_close_unref(data->bus);
    }
    if (data->pid) {
        D_DEBUG(data->pamh, "kill child thread:%ld!", data->pid);
        ret = pthread_cancel(data->pid);
        if (ESRCH != ret) {
            if (ret) {
                pam_syslog(data->pamh, LOG_ERR, "kill child thread error with: %s", strerror(ret));
            } else {
                ret = pthread_join(data->pid, NULL);
                if (ret && ESRCH != ret) {
                    pam_syslog(data->pamh, LOG_ERR, "join child thread error with: %s", strerror(ret));
                }
            }
        } // else 线程不存在，说明已经退出了

        data->pid = 0;
    }
    D_DEBUG(data->pamh, "free data!");
    free(data);

    if (auth_tok != NULL) {
        free(auth_tok);
        auth_tok = NULL;
    }

    if (term != NULL) {
        tcsetattr(STDIN_FILENO, TCSADRAIN, term);
        free(term);
        term = NULL;
    }
}

void sigterm_handler(int sig) {
    pam_clean_func(&userData);
    exit(-1);
}

static void load_user_locale(pam_handle_t *pamh, const char *username, const char *locale_path) {
    struct passwd *p;
    if ((p = getpwnam(username)) == NULL) {
        pam_syslog(pamh, LOG_WARNING, "run getpwnam failed: %s", strerror(errno));
        return;
    }

    size_t path_len = strlen(p->pw_dir) + strlen(locale_path) + 2;  // 2: '/' + '\0'
    if (path_len > PATH_MAX) {
        pam_syslog(pamh, LOG_WARNING, "user locale path exceeds PATH_MAX");
        return;
    }

    char *buff = malloc(path_len);
    strcpy(buff, p->pw_dir);
    strcat(buff, "/");
    strcat(buff, locale_path);

    FILE *f;
    if ((f = fopen(buff, "r")) == NULL) {
        D_DEBUG(pamh, "unable to open env file: %s: %s", buff, strerror(errno));
        free(buff);
        return;
    }

    char *pos = NULL;
    pam_syslog(pamh, LOG_INFO, "loading user locale");
    while (fgets(buff, MAX_BUF_SIZE, f) != NULL) {
        if ((pos = strchr(buff, '\n')) != NULL) {
            *pos = '\0';
        }

        pos = strchr(buff, '=');
        if (pos == NULL || buff == pos) {
            continue;
        }

        *pos = '\0';

        char *value = pos + 1;
        D_DEBUG(pamh, "setenv(%s, %s)", buff, value);
        setenv(buff, value, true);
    }

    fclose(f);
    free(buff);
    return;
}

static void *run_request_pw(void *user_data) {
    struct pam_response *rep = NULL;
    struct UserData *ud = user_data;

    D_DEBUG(ud->pamh, "[DEBUG] pam-module start get pass word thread: %ld", pthread_self());

    rep = send_msg(ud->pamh, "Password:", PAM_PROMPT_ECHO_OFF);
    if (rep) {
        dbus_method_setPassword(ud->pamh, ud->bus, ud->id, rep->resp);

        auth_tok = rep->resp;
        rep->resp = NULL;
        free((void *)rep);
        rep = NULL;
    }

    return NULL;
}

static void resolveVerifyMsg(const struct UserData *ud, const char *verify_msg, char *res_msg)
{
    int subcode = -1;
    int flag = -1;
    int code = -1;
    char subdata[MAX_BUF_SIZE] = {0};
    struct json_tokener *json_tor = json_tokener_new();

    D_DEBUG(ud->pamh, "verify msg :%s", verify_msg);
    do {
        struct json_object *parsed_json = NULL;
        struct json_object *json_flag = NULL;
        struct json_object *json_code = NULL;
        struct json_object *json_msg = NULL;
        enum json_tokener_error jerr;

        parsed_json = json_tokener_parse_ex(json_tor, verify_msg, (int)strlen(verify_msg));
        if ((jerr = json_tokener_get_error(json_tor)) != json_tokener_success) {
            pam_syslog(ud->pamh, LOG_ERR, "verify_msg json_tokener_parse error :%s", json_tokener_error_desc(jerr));
            break;
        }

        json_object_object_get_ex(parsed_json, "flag", &json_flag);
        json_object_object_get_ex(parsed_json, "code", &json_code);
        json_object_object_get_ex(parsed_json, "msg", &json_msg);

        flag = json_object_get_int(json_flag);
        code = json_object_get_int(json_code);
        sprintf(subdata, "%s", json_object_get_string(json_msg));

        switch (flag) {
        case AT_Fingerprint:{
            switch(code) {
            case 0: //Match 匹配, 之后结束认证。
                break;
            case 1: //NoMatch 不匹配，之后结束认证。
                break;
            case 2: //Error 失败，之后结束认证。
                switch (subcode) {
                case 1:
                    sprintf(res_msg, gettext("Verification error"));
                    send_msg(ud->pamh, res_msg, PAM_ERROR_MSG);
                    break;
                case 2:
                    sprintf(res_msg, gettext("Fingerprint verification unavailable, please input password"));
                    send_msg(ud->pamh, res_msg, PAM_ERROR_MSG);
                    break;
                }
                break;
            case 3: //Retry 重试，之后继续认证。
            {
                struct json_object *json_subcode = NULL;

                json_subcode = json_tokener_parse_ex(json_tor, subdata, (int)strlen(subdata));
                if ((jerr = json_tokener_get_error(json_tor)) != json_tokener_success) {
                    pam_syslog(ud->pamh, LOG_ERR, "subdata json_tokener_parse error :%s", json_tokener_error_desc(jerr));
                    break;
                }

                json_object_object_get_ex(json_subcode, "subcode", &json_subcode);
                subcode = json_object_get_int(json_subcode);

                switch(subcode) {
                case 1: //swipe-too-short 滑动太短
                    sprintf(res_msg, gettext("Clean your finger and try again"));
                    send_msg(ud->pamh, res_msg, PAM_ERROR_MSG);
                    break;
                case 2: //finger-not-centered 手指不在中间
                    sprintf(res_msg, gettext("Finger not centered"));
                    send_msg(ud->pamh, res_msg, PAM_ERROR_MSG);
                    break;
                case 3: //remove-and-retry 拿开手指再重新扫描
                    sprintf(res_msg, gettext("Clean your finger and try again"));
                    send_msg(ud->pamh, res_msg, PAM_ERROR_MSG);
                    break;
                case 4: //quality-bad 图像质量差
                    sprintf(res_msg, gettext("Unclear fingerprint, please clean your finger and try again"));
                    send_msg(ud->pamh, res_msg, PAM_ERROR_MSG);
                    break;
                case 5: //moved-too-fast 接触时间短
                    sprintf(res_msg, gettext("Finger moved too fast, please do not lift until prompted"));
                    send_msg(ud->pamh, res_msg, PAM_ERROR_MSG);
                    break;
                default:
                    pam_syslog(ud->pamh, LOG_ERR, "get error msg: %d", subcode);
                    break;
                }
            }
                break;
            case 4: //Disconnected 与设备失联，之后不要对设备进行任何操作。
                break;
            case 5: // TakenByOthers
                sprintf(res_msg, gettext("Password"));
                send_msg(ud->pamh, res_msg, PAM_TEXT_INFO);
                break;
            }
        }
            break;
        }
    } while(0);
    json_tokener_free(json_tor);
}

static void get_limits_info(const struct UserData *ud)
{
    char buff[LIMITS_BUF_SIZE] = {0};
    struct json_tokener *json_tor = json_tokener_new();

    do {
        struct json_object *parsed_json = NULL;
        struct json_object *json_flag = NULL;
        struct json_object *json_maxTries = NULL;
        struct json_object *json_numFailures = NULL;
        struct json_object *json_locked = NULL;
        struct json_object *json_unlockTime = NULL;
        enum json_tokener_error jerr;
        int flag = 0;

        if (dbus_method_get_limits(ud->pamh, ud->bus, ud->username, buff)) {
            pam_syslog(ud->pamh, LOG_ERR, "get limits failed");
            break;
        }

        parsed_json = json_tokener_parse_ex(json_tor, buff, (int)strlen(buff));
        if ((jerr = json_tokener_get_error(json_tor)) != json_tokener_success) {
            pam_syslog(ud->pamh, LOG_ERR, "verify_msg json_tokener_parse error :%s", json_tokener_error_desc(jerr));
            break;
        }

        int len = json_object_array_length(parsed_json);
        for (int i = 0; i < len; i++) {
            struct json_object *jobj = json_object_array_get_idx(parsed_json, i);

            json_object_object_get_ex(jobj, "flag", &json_flag);
            json_object_object_get_ex(jobj, "maxTries", &json_maxTries);
            json_object_object_get_ex(jobj, "numFailures", &json_numFailures);
            json_object_object_get_ex(jobj, "locked", &json_locked);
            json_object_object_get_ex(jobj, "unlockTime", &json_unlockTime);
         
            flag = json_object_get_int(json_flag);
            struct Limit *limit = NULL;
            switch (flag) {
            case AT_Password:
                limit = &limits[PASSWORD_LIMIT_INDEX];
                break;
            case AT_Fingerprint:
                limit = &limits[FINGERPRINT_LIMIT_INDEX];
                break;
            case AT_Face:
                limit = &limits[FACE_LIMIT_INDEX];
                break;
            case AT_ActiveDirectory:
                limit = &limits[ACTIVE_DIRECTORY_LIMIT_INDEX];
                break;
            }

            if (limit) {
                limit->maxTries = json_object_get_int(json_maxTries);
                limit->numFailures = json_object_get_int(json_numFailures);
                limit->locked = json_object_get_boolean(json_locked);
                strcpy(limit->unlockTime, json_object_get_string(json_unlockTime));
            }
        }
        //if (!jsonRes) break;
    } while (0);
    json_tokener_free(json_tor);
    //g_object_unref(parser);
}

static int bus_signal_cb(sd_bus_message *m, void *user_data, sd_bus_error *ret_error)
{
    UNUSED_VALUE(ret_error);
    int ret = 0;
    int code = -1;
    const char *signalmsg = NULL;
    const char *id = NULL;
    struct UserData *ud = user_data;
    char msg[MAX_BUF_SIZE] = {0};

    do {
        D_DEBUG(ud->pamh, "[DEBUG] signal_cb func be called");
        ret = sd_bus_message_read(m, "sis", &id, &code, &signalmsg);
        if (ret < 0) {
            D_DEBUG(ud->pamh, "[DEBUG] signal callback error :%s", strerror(errno));
            break;
        }

        D_DEBUG(ud->pamh, "[DEBUG] get signal data, id: %s, code: %d, signal msg: %s", id, code, signalmsg);

        if (strcmp(id, ud->id)) {
            D_DEBUG(ud->pamh, "[DEBUG] id not match signal id: %s, user data id: %s", id, ud->id);
            break;
        }
        switch(code) {
        case ASC_Success:
            g_res = PAM_SUCCESS;
            ud->needCancel = false;

            sprintf(msg, gettext("Verification successful"));
            send_msg(ud->pamh, msg, PAM_TEXT_INFO);

            if (auth_tok != NULL) {
                pam_set_item(ud->pamh, PAM_AUTHTOK, auth_tok);
            }

            break;
        case ASC_Failure: {
            g_res = PAM_AUTH_ERR;
            ud->needCancel = false;

            if (strcmp(signalmsg, "fingerprint") == 0) {
                get_limits_info(ud);

                struct Limit *limit = &limits[FINGERPRINT_LIMIT_INDEX];
                if (limit->locked) {
                    snprintf(msg, MAX_BUF_SIZE, gettext("Fingerprint locked, use password please"));
                } else {
                    int times = limit->maxTries - limit->numFailures;
                    if (times > 1) {
                        snprintf(msg, MAX_BUF_SIZE, gettext("Verification failed, %d chances left"), times);
                    } else {
                        snprintf(msg, MAX_BUF_SIZE, gettext("Verification failed, 1 chance left"));
                    }
                }

                send_msg(ud->pamh, msg, PAM_ERROR_MSG);
            }

            break;
        }
        case ASC_Cancel:
        case ASC_Timeout:
        case ASC_Error:
            ud->needCancel = false;
            D_DEBUG(ud->pamh, "[DEBUG] set result code: %d", code);
            break;
        case ASC_Verify:
            D_DEBUG(ud->pamh, "start resolve verify msg: %s", signalmsg);
            memset(msg, 0, MAX_BUF_SIZE);
            resolveVerifyMsg(ud, signalmsg, msg);
            break;
        }
    } while (0);

    return 0;
}

PAM_EXTERN int pam_sm_authenticate(pam_handle_t *pamh, int flags, int argc, const char **argv)
{
    UNUSED_VALUE(flags);
    userData = NULL;
    const char *username = NULL;
    sd_bus *bus = NULL;
    int ret = 0;
    int idx = 0;
    int authflag = AT_Password | AT_Fingerprint;
    int timeout = 60000;
    char buff[MAX_BUF_SIZE] = {0};
    char user_locale[MAX_BUF_SIZE] = {0};
    int reqid = -1;

    for (idx = 0; idx < argc; ++idx) {
        const char *splitch = strchr(argv[idx], '=');

        strncpy(buff, argv[idx], MAX_BUF_SIZE - 1);
        if (splitch != NULL) {
            buff[((size_t)(splitch - argv[idx]))/sizeof(char)] = '\0';

            if (!strcmp(buff, TIMEOUT_FLAG)) {
                timeout = atoi(splitch + 1);
                continue;
            }

            if (!strcmp(buff, USER_LOCALE_FLAG)) {
                strncpy(user_locale, splitch + 1, MAX_BUF_SIZE - 1);
                continue;
            }
        }

        if (!strcmp(buff, DEBUG_FLAG)) {
            g_beDebug = true;
        }
    }

    term = malloc(sizeof(struct termios));
    ret = tcgetattr(STDIN_FILENO, term);
    if (ret == -1) {
        pam_syslog(pamh, LOG_ERR, "Cannot get original tty attributes: %s", strerror(errno));
        free(term);
        term = NULL;
    }

    userData = (struct UserData *)malloc(sizeof(struct UserData));

    struct sigaction act;
    sigemptyset(&act.sa_mask);
    act.sa_flags = 0;
    act.sa_handler = &sigterm_handler;
    if (sigaction(SIGTERM, &act, NULL) < 0) {
        pam_syslog(pamh, LOG_ERR, "cannot set SIGTERM handler: %s", strerror(errno));
        return PAM_ABORT;
    }

    pam_get_user(pamh, &username, "Username: ");
    D_DEBUG(pamh, "[DEBUG] start auth with username:---%s---", username);

    if (user_locale[0] != '\0') {
        load_user_locale(pamh, username, user_locale);
    }

    setlocale(LC_ALL, "");
    textdomain("deepin-authentication");

    g_res = -1;
    pthread_cleanup_push(pam_clean_func, &userData);

    do {
        ret = sd_bus_open_system(&bus);
        if (ret < 0) {
            pam_syslog(pamh, LOG_ERR, "Failed to connect to system bus: %s", strerror(-ret));
            break;
        }

        // 尝试一键登录，不处理失败，失败后走正常的验证逻辑
        dbus_method_oneKeyLogin(pamh, bus, username, buff);

        // 一键登录没有返回信息，说明失败或该账户没有指纹，使用正常的验证逻辑
        if (strlen(buff) == 0) {
            if (dbus_method_authenticate(pamh, bus, username, authflag, timeout, buff)) {
                g_res = PAM_ABORT;
                break;
            }
        }

        sscanf(buff, "{\"id\":\"%d\",\"flag\":%d}", &reqid, &authflag);

        userData->pamh = pamh;
        userData->bus = bus;
        userData->pid = 0;
        userData->needCancel = true;
        sprintf(userData->id, "%d", reqid);
        sprintf(userData->username, "%s", username);

        D_DEBUG(pamh, "[DEBUG] start authenticate id: %s, username: %s!", userData->id, userData->username);

        ret = sd_bus_match_signal(bus, NULL, DBUS_SERVICE, DBUS_PATH, DBUS_INTERFACE,
                                  AUTH_SIGNAL_STATUS, bus_signal_cb, userData);
        if (ret < 0) {
            pam_syslog(pamh, LOG_ERR, "add match error!");
            g_res = PAM_ABORT;
            break;
        }

        // 发送开启的认证方式
        if (authflag & AT_Fingerprint) {
            snprintf(buff, MAX_BUF_SIZE, gettext("Verify your fingerprint or password"));
            send_msg(pamh, buff, PAM_TEXT_INFO);
        } else {
            snprintf(buff, MAX_BUF_SIZE, gettext("Password"));
            send_msg(pamh, buff, PAM_TEXT_INFO);
        }

        // 如果开启了密码认证，则向app请求密码数据
        if(authflag & AT_Password) {
            int err = 0;

            D_DEBUG(pamh, "[DEBUG] start password authenticate!");

            err = pthread_create(&userData->pid, NULL, run_request_pw, userData);
            if (err != 0) {
                pam_syslog(pamh, LOG_ERR, "create password thread error: %d", err);
                g_res = PAM_ABORT;
                break;
            }

            D_DEBUG(pamh, "create password thread id: %ld", userData->pid);
        }

        D_DEBUG(pamh, "start dbus loop, g_res: %d", g_res);

        while(g_res == -1) {
            sd_bus_message *m = NULL;
            ret = sd_bus_process(bus, &m);
            if (ret < 0) {
                pam_syslog(pamh, LOG_ERR, "sd_bus_process failed: %s", strerror(ret));
                g_res = PAM_ABORT;
                break;
            }

            if (ret > 0) {
                continue;
            }

            ret = sd_bus_wait(bus, (uint64_t) 100000);
            if (ret < 0) {
                pam_syslog(pamh, LOG_ERR, "sd_bus_wait failed: %s", strerror(ret));
                g_res = PAM_ABORT;
                break;
            }
        }
    } while(0);

    pthread_cleanup_pop(1);

    D_DEBUG(pamh, "auth result: %d", g_res);
    return g_res;
}

static int
dbus_method_authenticate(pam_handle_t *pamh, sd_bus *bus,
                         const char *username, int flags, int timeout, char *id)
{
    sd_bus_error err = SD_BUS_ERROR_NULL;
    sd_bus_message *reply = NULL;
    char *idreq = NULL;
    int ret = 0;

    do{
        ret = sd_bus_call_method(bus, DBUS_SERVICE, DBUS_PATH, DBUS_INTERFACE,
                                 "Authenticate", &err, &reply, "sii", username, flags, timeout);
        if (ret < 0) {
            pam_syslog(pamh, LOG_ERR, "Failed to call 'Authenticate': %s, %s",
                     err.name, err.message);
            break;
        }

        ret = sd_bus_message_read(reply, "s", &idreq);
        if (ret < 0) {
            D_DEBUG(pamh, "Failed to read 'Authenticate' value: %s",
                    strerror(errno));
            break;
        }

        D_DEBUG(pamh, "[DEBUG] Authenticate return id: %s", idreq);
        sprintf(id, "%s", idreq);
    } while(0);

    sd_bus_error_free(&err);
    sd_bus_message_unref(reply);

    return ret < 0 ? 1 : 0;
}

static int
dbus_method_get_limits(pam_handle_t *pamh, sd_bus *bus, const char *username, char *limits)
{
    sd_bus_error err = SD_BUS_ERROR_NULL;
    sd_bus_message *reply = NULL;
    char *res = NULL;
    int ret = 0;

    do {
        ret = sd_bus_call_method(bus, DBUS_SERVICE, DBUS_PATH, DBUS_INTERFACE, "GetLimits", &err, &reply, "s", username);
        if (ret < 0) {
            pam_syslog(pamh, LOG_ERR, "Failed to call 'GetLimits': %s, %s",
                       err.name, err.message);
            break;
        }

        ret = sd_bus_message_read(reply, "s", &res);
        if (ret < 0) {
            D_DEBUG(pamh, "Failed to read 'GetLimits' value: %s",
                       strerror(errno));
            break;
        }

        D_DEBUG(pamh, "[DEBUG] GetLimits return: %s", res);
        sprintf(limits, "%s", res);
    } while (0);

    sd_bus_error_free(&err);
    sd_bus_message_unref(reply);

    return ret < 0 ? 1 : 0;
}

static int
dbus_method_cancelAuthenticate(pam_handle_t *pamh, sd_bus *bus,
                               const char *id)
{
    sd_bus_error err = SD_BUS_ERROR_NULL;
    sd_bus_message *reply = NULL;
    int ret = 0;

    do {
        ret = sd_bus_call_method(bus, DBUS_SERVICE, DBUS_PATH, DBUS_INTERFACE,
                                 "CancelAuthenticate", &err, &reply, "s", id);
        if (ret < 0) {
            pam_syslog(pamh, LOG_ERR, "Failed to call 'CancelAuthenticate': %s, %s",
                       err.name, err.message);
            break;
        }

        D_DEBUG(pamh, "[DEBUG] call CancelAuthenticate finished");
    } while(0);

    sd_bus_error_free(&err);
    sd_bus_message_unref(reply);

    return ret < 0 ? 1 : 0;
}

static int call_setPassword_cb(sd_bus_message *m, void *userdata, sd_bus_error *ret_error) {
    UNUSED_VALUE(m);
    UNUSED_VALUE(userdata);
    UNUSED_VALUE(ret_error);
    return 0;
}

static int
dbus_method_setPassword(pam_handle_t *pamh, sd_bus *bus,
                        const char *id,const char *password)
{
    sd_bus_error err = SD_BUS_ERROR_NULL;
    sd_bus_message *reply = NULL;
    sd_bus_slot *slot = NULL;
    int ret = 0;

    do{
        D_DEBUG(pamh, "[DEBUG] start SetPassword with id: %s, password %s", id, password);

        ret = sd_bus_call_method_async(bus, &slot, DBUS_SERVICE, DBUS_PATH, DBUS_INTERFACE,
                                       "SetPassword", &call_setPassword_cb, NULL, "ss", id, password);
        if (ret < 0) {
            pam_syslog(pamh, LOG_ERR, "Failed to call 'SetPassword': %s, %s",
                       err.name, err.message);
            break;
        }

        D_DEBUG(pamh, "[DEBUG] call SetPassword finished");
    } while(0);

    sd_bus_error_free(&err);
    sd_bus_message_unref(reply);

    return ret < 0 ? 1 : 0;
}

static int
dbus_method_oneKeyLogin(pam_handle_t *pamh, sd_bus *bus, const char *username, char *id)
{
    sd_bus_error err = SD_BUS_ERROR_NULL;
    sd_bus_message *reply = NULL;
    int ret = 0;
    char *idreq = NULL;

    do {
        D_DEBUG(pamh, "[DEBUG] start OneKeyLogin");

        ret = sd_bus_call_method(bus, DBUS_SERVICE, DBUS_PATH, DBUS_INTERFACE, "OneKeyLogin", &err, &reply, "s", username);
        if (ret < 0) {
            pam_syslog(pamh, LOG_ERR,
                       "Failed to call 'OneKeyLogin': %s, %s",
                       err.name, err.message);
            break;
        }

        ret = sd_bus_message_read(reply, "s", &idreq);
        if (ret < 0) {
            D_DEBUG(pamh, "Failed to read 'Authenticate' value: %s",
                    strerror(errno));
            break;
        }

        D_DEBUG(pamh, "[DEBUG] OneKeyLogin return id: %s", idreq);
        sprintf(id, "%s", idreq);
    } while (0);

    sd_bus_error_free(&err);
    sd_bus_message_unref(reply);

    return ret < 0 ? 1 : 0;
}

static struct pam_response *
send_msg(pam_handle_t *pamh, const char *msg, int style)
{
    pam_syslog(pamh, LOG_INFO, "%s", msg);

    const struct pam_message pmsg = {
        .msg = msg,
        .msg_style = style,
    };
    const struct pam_message *pmsg_ptr = &pmsg;
    const struct pam_conv *pconv = NULL;
    struct pam_response *presp = NULL;
    int ret = 0;

    do {
        ret = pam_get_item(pamh, PAM_CONV, (const void**)&pconv);
        if (ret != PAM_SUCCESS) {
            pam_syslog(pamh, LOG_ERR, "pam module get conv item error: %s!", strerror(-ret));
            break;
        }

        if (!pconv || !pconv->conv) {
            pam_syslog(pamh, LOG_ERR, "pam module pconv or pconv->conv is nullptr, error!");
            break;
        }

        ret = pconv->conv(1, &pmsg_ptr, &presp, pconv->appdata_ptr);
        if (ret != PAM_SUCCESS) {
            pam_syslog(pamh, LOG_ERR, "in style %d Cannot get pam module conv : %s!",style, strerror(-ret));
            if (style == PAM_PROMPT_ECHO_OFF || style == PAM_PROMPT_ECHO_ON) {
                g_res = g_res == -1 ? PAM_ABORT : g_res;
            }
            break;
        }
    } while(0);

    return presp;
}

PAM_EXTERN int pam_sm_setcred(pam_handle_t *pamh, int flags,
                              int argc, const char **argv)
{
    return PAM_SUCCESS;
}
