package authenticate

import (
	"encoding/json"
	"errors"
	"sync"
	"time"

	authenticate "github.com/linuxdeepin/go-dbus-factory/com.deepin.daemon.authenticate"
	"pkg.deepin.io/dde/authentication/pkg/fingerprint"
	pkgfp "pkg.deepin.io/dde/authentication/pkg/fingerprint"
	"github.com/godbus/dbus"
	"pkg.deepin.io/lib/dbusutil/proxy"
)

var (
	claimed   *FingerprintTx
	claimedMu sync.Mutex
)

type FingerprintTx struct {
	baseTx
	quit    chan struct{}
	release chan struct{}

	fpObj *authenticate.Fingerprint

	isEnded bool
	endedMu sync.Mutex
}

func newFingerprintTx() *FingerprintTx {
	tx := &FingerprintTx{}
	tx.type0 = AuthTypeFingerprint
	return tx
}

func (ft *FingerprintTx) setPassword(password string) {
}

func (ft *FingerprintTx) end() {
	ft.endedMu.Lock()
	defer ft.endedMu.Unlock()

	logger.Debug(ft, "end")
	if ft.isEnded {
		return
	}

	close(ft.quit)

	// wait fingerprint device released
	<-ft.release

	logger.Debug(ft, "end return")
	ft.isEnded = true
	claimed = nil
}

func hasFingerprintDevice() (bool, error) {
	sysBus, err := dbus.SystemBus()
	if err != nil {
		return false, err
	}

	fpObj := authenticate.NewFingerprint(sysBus)
	defaultDevice, err := fpObj.DefaultDevice().Get(0)
	if err != nil {
		return false, err
	}
	if defaultDevice == "" {
		return false, nil
	}

	return true, nil
}

func cancelLastFingerprintTx() {
	claimedMu.Lock()
	defer claimedMu.Unlock()

	if claimed != nil {
		claimed.end()
	}
	claimed = nil
}

func (ft *FingerprintTx) checkCapability(cap int32) (bool, error) {
	defaultDevice, err := ft.fpObj.DefaultDevice().Get(0)
	if err != nil {
		return false, err
	}

	devicesStr, err := ft.fpObj.Devices().Get(0)
	if err != nil {
		return false, err
	}

	var devices []fingerprint.DeviceInfo
	err = json.Unmarshal([]byte(devicesStr), &devices)
	if err != nil {
		return false, err
	}

	for _, v := range devices {
		if v.Name == defaultDevice {
			if (v.Capability & cap) != 0 {
				return true, nil
			}
			break
		}
	}

	return false, nil
}

func (ft *FingerprintTx) initTx(username string) error {
	sysBus, err := dbus.SystemBus()
	if err != nil {
		logger.Debug("connect to system bus failed", err)
		return err
	}

	ft.fpObj = authenticate.NewFingerprint(sysBus)

	if username == pkgfp.EmptyUsername {
		canOneKeyLogin, err := ft.checkCapability(fingerprint.CapabilityOneKeyLogin)
		if err != nil {
			return err
		}
		if !canOneKeyLogin {
			return errors.New("the device does not support one-key login")
		}
	} else {
		fingers, err := ft.fpObj.ListFingers(0, username)
		if err != nil {
			return err
		}
		if len(fingers) == 0 {
			return errors.New("no fingers")
		}
	}

	claimedMu.Lock()
	if claimed != nil {
		claimed.end()
	}
	claimed = ft
	claimedMu.Unlock()

	ft.quit = make(chan struct{})
	ft.release = make(chan struct{})

	err = ft.fpObj.Claim(0, username, true)
	if err != nil {
		logger.Debug("claim fingerprint device failed", err)
		// close the ft.release,will not blocked in the func `end`
		close(ft.release)
		return err
	}

	return nil
}

func (ft *FingerprintTx) authenticate() error {
	username := ft.username()

	go func() {
		authResult := ft.verify()

		// release device
		err := ft.fpObj.Claim(0, username, false)
		if err != nil {
			logger.Warning(ft, err)
		}

		close(ft.release)

		if authResult != nil {
			ft.giveResult(ft, authResult)
		}
	}()
	return nil
}

type verifyResult struct {
	code int32
	msg  string
}

func (ft *FingerprintTx) verify() (result AuthResult) {
	ft.fpObj.InitSignalExt(ft.parent.m.sysSigLoop, true)

	verifyResultCh := make(chan verifyResult)

	_, err := ft.fpObj.ConnectVerifyStatus(func(username string, code int32, msg string) {
		if ft.username() == pkgfp.EmptyUsername {
			ft.parent.m.setOneKeyLoginResult(ft.parent.id, ft.type0, username, code == pkgfp.VerifyStatusMatch)
			return
		}

		if username != ft.username() {
			logger.Warningf("username %s != %s", username, ft.username())
			return
		}
		result := verifyResult{
			code: code,
			msg:  msg,
		}
		logger.Debug(ft, "signal VerifyStatus", result)
		ft.parent.m.emitSignalStatusVerify(ft.parent.id, AuthenticationFlagFingerprint, int(code), msg)
		select {
		case verifyResultCh <- result:
		case <-time.After(100 * time.Millisecond):
		}
	})
	if err != nil {
		logger.Warning(err)
	}

	result = ft.doVerify(verifyResultCh)
	logger.Debugf("%v doVerify result: %v", ft, result)

	ft.fpObj.RemoveHandler(proxy.RemoveAllHandlers)

	return result
}

func (ft *FingerprintTx) doVerify(verifyResultCh chan verifyResult) (result AuthResult) {
	logger.Debug(ft, "Verify start")
	err := ft.fpObj.Verify(0, "any")
	if err != nil {
		logger.Warning(err)
		result = &AuthResultError{err: err}
		return
	}

	var code int32 = -1
loop:
	for {
		select {
		//case <-time.After(10 * time.Second):
		//	logger.Warning("timed out")
		//	result = AuthResultTimeout{}
		//	break loop
		case <-ft.quit:
			break loop
		case result := <-verifyResultCh:
			done, err := pkgfp.IsVerifyStatusDone(int(result.code))
			if err != nil {
				logger.Warning(err)
			}
			logger.Debug(ft, "is verifyStatus done:", pkgfp.VerifyStatusToString(int(result.code)), done)
			if done {
				code = result.code
				break loop
			}
		}
	}

	logger.Debug(ft, "stop verify")
	err = ft.fpObj.StopVerify(0)
	if err != nil {
		logger.Warning(ft, "call StopVerify err:", err)
		result = &AuthResultError{err: err}
		return
	}

	switch code {
	case -1:
		if !ft.parent.isEnded() {
			ft.parent.m.emitSignalStatusVerify(ft.parent.id, AuthenticationFlagFingerprint, pkgfp.VerifyStatusTakenByOthers, "")
		}
	case pkgfp.VerifyStatusMatch:
		result = &AuthResultSuccess{}
	case pkgfp.VerifyStatusNoMatch:
		result = &AuthResultFailure{}
	case pkgfp.VerifyStatusError:
		logger.Warning(ft, "verify error")
		result = &AuthResultError{err: errors.New("verify error")}
	}
	return
}
