package authenticate

import (
	"sync"
)

type Group struct {
	m        *Manager
	id       string
	username string
	txs      []Tx
	ended    bool
	mu       sync.Mutex
	resultCh chan Tx
}

func (g *Group) getKeyboardResult() AuthResult {
	var results AuthResults
	for _, tx := range g.txs {
		switch tx.getType() {
		case AuthTypePassword, AuthTypeActiveDirectory:
			results = append(results, tx.getResult())
		}
	}
	logger.Debugf("auth group %v getKeyboardResult results: %v", g.id, results)
	if len(results) == 0 {
		return nil
	}
	success := results.getFirstSuccess()
	if success != nil {
		return success
	}

	if len(results) > 1 {
		if results.allFailure() {
			logger.Debugf("auth group %v all failure", g.id)
			result := &AuthResultFailure{}
			result.setType("keyboard")
			return result
		}
		return nil
	}
	return results[0]
}

type AuthResults []AuthResult

func (results AuthResults) getFirstSuccess() AuthResult {
	for _, result := range results {
		switch result.(type) {
		case *AuthResultSuccess:
			return result
		}
	}
	return nil
}

func (results AuthResults) getFirstFailure() AuthResult {
	for _, result := range results {
		switch result.(type) {
		case *AuthResultFailure:
			return result
		}
	}
	return nil
}

func (results AuthResults) allFailure() bool {
	count := 0
	for _, result := range results {
		switch result.(type) {
		case *AuthResultFailure:
			count++
		}
	}
	return count == len(results)
}

func (g *Group) getFingerprintResult() AuthResult {
	for _, tx := range g.txs {
		switch tx.getType() {
		case AuthTypeFingerprint:
			return tx.getResult()
		}
	}
	return nil
}

func (g *Group) getFinalResult() AuthResult {
	kbd := g.getKeyboardResult()
	fp := g.getFingerprintResult()
	logger.Debugf("auth group %v kbd: %v", g.id, kbd)
	logger.Debugf("auth group %v fp: %v", g.id, fp)
	results := AuthResults{kbd, fp}
	failure := results.getFirstFailure()
	if failure != nil {
		return failure
	}
	success := results.getFirstSuccess()
	if success != nil {
		return success
	}

	if g.allTxsStopped() {
		logger.Debugf("auth group %v all txs stopped", g.id)
		return &AuthResultFailure{}
	}

	return nil
}

func newGroup(m *Manager, id, username string) *Group {
	return &Group{
		m:        m,
		id:       id,
		username: username,
	}
}

func (g *Group) addTx(tx Tx) {
	tx.init(g)
	g.txs = append(g.txs, tx)
}

func (g *Group) isEnded() bool {
	g.mu.Lock()
	v := g.ended
	g.mu.Unlock()
	return v
}

func (g *Group) getAndMarkEnded() bool {
	g.mu.Lock()
	v := g.ended
	g.ended = true
	g.mu.Unlock()
	return v
}

func (g *Group) listenResultCh() {
	for {
		tx, ok := <-g.resultCh
		if !ok {
			logger.Debugf("auth group %v result ch closed", g.id)
			return
		}
		logger.Debugf("auth group %v receive result, type: %v", g.id, tx.getType())

		switch tx0 := tx.(type) {
		case *FingerprintTx:
			result := tx0.getResult()
			switch result.(type) {
			case *AuthResultTimeout:
				g.m.emitSignalStatus(g.id, StatusCodeTimeout, tx0.getType())
			}
		}
		result := g.getFinalResult()
		logger.Debugf("auth group %v final result: %v", g.id, result)
		if result != nil {
			statusCode := result.StatusCode()
			g.end(statusCode, result)
			g.m.emitSignalStatus(g.id, statusCode, result.getType())
		}
	}
}

func (g *Group) allTxsStopped() bool {
	// 不用 g.mu 加锁
	count := 0
	for _, tx := range g.txs {
		if tx.getResult() != nil {
			count++
		}
	}
	return len(g.txs) == count
}

func (g *Group) setPassword(password string) {
	for _, tx := range g.txs {
		tx.setPassword(password)
	}
}

func (g *Group) cancel() {
	logger.Debugf("group %s canceled", g.id)
	g.end(StatusCodeCancel, nil)
	g.m.emitSignalStatus(g.id, StatusCodeCancel, "")
}

func (g *Group) end(statusCode int, result AuthResult) {
	ended := g.getAndMarkEnded()
	if ended {
		return
	}

	logger.Debugf("auth group %v end, statusCode: %s, result: %v", g.id, statusCodeToString(statusCode), result)
	close(g.resultCh)
	for _, tx := range g.txs {
		tx.end()
	}
	g.m.removeGroup(g.id)
	if statusCode == StatusCodeFailure {
		// 结算失败
		type0 := result.getType()
		if type0 == "keyboard" {
			for _, tx := range g.txs {
				txType := tx.getType()
				switch txType {
				case AuthTypePassword, AuthTypeActiveDirectory:
					g.m.fail(g.username, txType)
				}
			}
		} else if type0 != "" {
			g.m.fail(g.username, type0)
		}
	} else if statusCode == StatusCodeSuccess {
		g.m.resetLimits(g.username)
	}
}

func (g *Group) authenticate() {
	g.resultCh = make(chan Tx)
	go g.listenResultCh()
	for _, tx := range g.txs {
		err := tx.authenticate()
		if err != nil {
			logger.Warning(err)
		}
	}
}
