package authenticate

import (
	"encoding/json"
	"io/ioutil"
	"time"
)

const (
	configFile      = "/var/lib/deepin/authenticate/config.json"
	limitStatesFile = "/var/lib/deepin/authenticate/limit-states.json"
)

type Config struct {
	Limits []*LimitConfig
}

func loadConfig(filename string, cfg *Config) error {
	data, err := ioutil.ReadFile(filename)
	if err != nil {
		return err
	}
	err = json.Unmarshal(data, cfg)

	// 修正配置
	// 当 MaxTries < 0 时，修正为 0，不限制。
	for _, limitConfig := range cfg.Limits {
		if limitConfig.MaxTries < 0 {
			limitConfig.MaxTries = 0
		}
	}

	return err
}

type LimitConfig struct {
	Type       string
	UnlockSecs int // 超出认证失败次数后，解除锁定需要等待的时间（秒数），如果 <0 则永久锁定。
	MaxTries   int // 认证失败几次后拒绝, 如果 == 0 则不限制
}

type LimitState struct {
	Type        string
	NumFailures int
	LockAt      time.Time
}

type LimitForExport struct {
	Type string `json:"type"`
	Flag int    `json:"flag"`

	UnlockSecs int `json:"unlockSecs"`
	MaxTries   int `json:"maxTries"`

	NumFailures int       `json:"numFailures"`
	Locked      bool      `json:"locked"`
	UnlockTime  time.Time `json:"unlockTime"`
}

func loadLimitStates(filename string, states *map[string][]LimitState) error {
	data, err := ioutil.ReadFile(filename)
	if err != nil {
		return err
	}
	err = json.Unmarshal(data, states)
	return err
}

func saveLimitStates(filename string, states map[string][]LimitState) error {
	data, err := json.Marshal(states)
	if err != nil {
		return err
	}
	err = ioutil.WriteFile(filename, data, 0644)
	return err
}

type limit struct {
	type0       string
	flag        int
	numFailures int
	lockAt      time.Time
	lastFailAt  time.Time // 最后一次失败的时间，暂时无用
	unlockSecs  int
	maxTries    int
}

func (l *limit) tryReset(now time.Time) {
	if l.unlockSecs < 0 {
		// 不能自动解锁
		return
	}
	if l.numFailures == l.maxTries && l.maxTries > 0 {
		unlockDuration := time.Second * time.Duration(l.unlockSecs)
		if now.Sub(l.lockAt) > unlockDuration {
			// 可以重置了
			l.reset()
		}
	}
}

func (l *limit) reset() {
	l.numFailures = 0
	l.lockAt = time.Time{}
}

func (l *limit) allow(now time.Time) (bool, time.Time) {
	if l.maxTries == 0 {
		// 不受限制
		return true, time.Time{}
	}
	//logger.Debugf("l: %#v", l)
	l.tryReset(now)
	if l.numFailures < l.maxTries {
		return true, time.Time{}
	}
	if l.unlockSecs < 0 {
		return false, time.Time{}
	}
	unlockDuration := time.Second * time.Duration(l.unlockSecs)
	return false, l.lockAt.Add(unlockDuration)
}

func (l *limit) fail(now time.Time) {
	if l.maxTries == 0 {
		return
	}
	l.tryReset(now)
	l.lastFailAt = now
	l.numFailures++
	if l.numFailures == l.maxTries {
		l.lockAt = now
	}
}
