package authenticate

import (
	"errors"
	"fmt"
	"os"
	"path/filepath"
	"strconv"
	"sync"
	"time"

	authenticate "github.com/linuxdeepin/go-dbus-factory/com.deepin.daemon.authenticate"
	ofdbus "github.com/linuxdeepin/go-dbus-factory/org.freedesktop.dbus"
	"pkg.deepin.io/dde/authentication/pkg/common"
	pkgfp "pkg.deepin.io/dde/authentication/pkg/fingerprint"
	"github.com/godbus/dbus"
	"pkg.deepin.io/lib/dbusutil"
)

const (
	dbusServicePath      = "/com/deepin/daemon/Authenticate"
	dbusServiceIntreface = "com.deepin.daemon.Authenticate"
)

const (
	AuthenticationFlagPassword        = 1 << iota // 启用密码认证
	AuthenticationFlagFingerprint                 // 启用指纹认证
	AuthenticationFlagFace                        // 启用人脸认证
	AuthenticationFlagActiveDirectory             // 启用 AD 域认证
)

const (
	AuthTypePassword        = "password"
	AuthTypeFingerprint     = "fingerprint"
	AuthTypeFace            = "face"
	AuthTypeActiveDirectory = "active directory"
)

func authTypeToFlag(type0 string) int {
	switch type0 {
	case AuthTypePassword:
		return AuthenticationFlagPassword

	case AuthTypeFingerprint:
		return AuthenticationFlagFingerprint

	case AuthTypeFace:
		return AuthenticationFlagFace

	case AuthTypeActiveDirectory:
		return AuthenticationFlagActiveDirectory

	default:
		return 0
	}
}

const (
	StatusCodeSuccess = iota // 成功
	StatusCodeFailure        // 失败
	StatusCodeCancel         // 取消
	StatusCodeTimeout        // 超时
	StatusCodeError          // 错误
	StatusCodeVerify         // 验证中
)

func statusCodeToString(code int) string {
	switch code {
	case StatusCodeSuccess:
		return "success"
	case StatusCodeFailure:
		return "failure"
	case StatusCodeCancel:
		return "cancel"
	case StatusCodeTimeout:
		return "timeout"
	case StatusCodeError:
		return "error"
	case StatusCodeVerify:
		return "verify"
	default:
		return fmt.Sprintf("invalid-status-code(%d)", code)
	}
}

type oneKeyLoginResult struct {
	gid      string
	txType   string
	flag     int
	username string
	result   bool
}

type authenticateReturn struct {
	Id   string `json:"id"`
	Flag int    `json:"flag"`
}

type Manager struct {
	nextId   uint
	nextIdMu sync.Mutex

	config Config

	groups   map[string]*Group
	groupsMu sync.Mutex

	limits       map[string]map[string]*limit
	limitsMu     sync.Mutex
	saveLimitsMu sync.Mutex

	service    *dbusutil.Service
	sysSigLoop *dbusutil.SignalLoop
	dbusDaemon *ofdbus.DBus

	SupportedFlags int32

	oneKeyLoginRes    oneKeyLoginResult
	oneKeyLoginWaitCh chan oneKeyLoginResult

	// nolint
	methods *struct {
		Authenticate       func() `in:"username,flag,timeout" out:"result"`
		CancelAuthenticate func() `in:"id"`
		PreOneKeyLogin     func() `in:"flag" out:"result"`
		OneKeyLogin        func() `in:"username" out:"result"`
		SetPassword        func() `in:"id,password"`
		GetLimits          func() `in:"username" out:"limits"`
	}

	// nolint
	signals *struct {
		Status struct {
			id   string
			code int32
			msg  string
		}
	}
}

var allAuthTypes = []string{AuthTypePassword, AuthTypeActiveDirectory, AuthTypeFingerprint, AuthTypeFace}

func (m *Manager) initLimits() {
	m.limits = make(map[string]map[string]*limit)
	m.loadLimitStates()
}

func (m *Manager) initLimitsByUsername(username string) {
	_, ok := m.limits[username]
	if ok {
		return
	}

	m.limits[username] = make(map[string]*limit)
	for _, authType := range allAuthTypes {
		m.limits[username][authType] = nil
	}

	for _, limitConfig := range m.config.Limits {
		type0 := limitConfig.Type
		flag := authTypeToFlag(type0)
		if flag == 0 {
			logger.Warningf("invalid auth type %q in limit config", type0)
			continue
		}

		m.limits[username][type0] = &limit{
			type0:      type0,
			flag:       flag,
			unlockSecs: limitConfig.UnlockSecs,
			maxTries:   limitConfig.MaxTries,
		}
	}

	for _, authType := range allAuthTypes {
		if m.limits[username][authType] == nil {
			logger.Debugf("auth type %q use default config", authType)
			m.limits[username][authType] = &limit{
				type0:      authType,
				flag:       authTypeToFlag(authType),
				unlockSecs: 0,
				maxTries:   0,
			}
		}
	}
}

func (m *Manager) allowTx(username, type0 string) bool {
	l := m.limits[username][type0]
	if l == nil {
		logger.Warning("not found limit for", type0)
		return false
	}

	now := time.Now()
	m.limitsMu.Lock()
	allow, _ := l.allow(now)
	m.limitsMu.Unlock()
	return allow
}

func (m *Manager) fail(username, type0 string) {
	l := m.limits[username][type0]
	if l == nil {
		logger.Warning("not found limit for", type0)
		return
	}
	logger.Debug("fail", type0)
	now := time.Now()

	m.limitsMu.Lock()
	l.fail(now)
	states := m.getLimitStates()
	m.limitsMu.Unlock()

	m.saveLimitStates(states)
}

func (m *Manager) resetLimits(username string) {
	m.limitsMu.Lock()
	for _, ll := range m.limits[username] {
		ll.reset()
	}
	states := m.getLimitStates()
	m.limitsMu.Unlock()

	m.saveLimitStates(states)
}

func (m *Manager) loadLimitStates() {
	var states map[string][]LimitState
	err := loadLimitStates(limitStatesFile, &states)
	if err != nil {
		if !os.IsNotExist(err) {
			logger.Warning(err)
		}
	}

	for username, v := range states {
		m.initLimitsByUsername(username)
		for _, state := range v {
			ll := m.limits[username][state.Type]
			if ll == nil {
				continue
			}
			ll.lockAt = state.LockAt
			ll.numFailures = state.NumFailures
		}
	}
}

func (m *Manager) getLimitStates() map[string][]LimitState {
	states := make(map[string][]LimitState)
	for username, v := range m.limits {
		for _, ll := range v {
			states[username] = append(states[username], LimitState{
				Type:        ll.type0,
				NumFailures: ll.numFailures,
				LockAt:      ll.lockAt,
			})
		}
	}
	return states
}

func (m *Manager) saveLimitStates(states map[string][]LimitState) {
	logger.Debugf("saveLimitStates: %#v", states)
	m.saveLimitsMu.Lock()
	defer m.saveLimitsMu.Unlock()

	dir := filepath.Dir(limitStatesFile)
	err := os.MkdirAll(dir, 0755)
	if err != nil {
		logger.Warning(err)
		return
	}

	err = saveLimitStates(limitStatesFile, states)
	if err != nil {
		logger.Warning(err)
	}
}

func (m *Manager) getLimits(username string) ([]LimitForExport, error) {
	m.limitsMu.Lock()
	defer m.limitsMu.Unlock()

	result := make([]LimitForExport, 0)
	limits, ok := m.limits[username]
	if !ok {
		return result, nil
	}

	now := time.Now()
	for _, l := range limits {
		allow, unlockTime := l.allow(now)
		le := LimitForExport{
			Type:        l.type0,
			UnlockSecs:  l.unlockSecs,
			MaxTries:    l.maxTries,
			Flag:        l.flag,
			NumFailures: l.numFailures,
			Locked:      !allow,
			UnlockTime:  unlockTime,
		}

		result = append(result, le)
	}
	return result, nil
}

func (m *Manager) emitSignalStatus(id string, code int, msg string) {
	logger.Debugf("emit signal Status id: %s, code: %s, msg: %s", id, statusCodeToString(code), msg)
	err := m.service.Emit(m, "Status", id, int32(code), msg)
	if err != nil {
		logger.Warning(err)
	}
}

func (m *Manager) emitSignalStatusVerify(id string, authFlag int, code int, msg string) {
	m.emitSignalStatusJSON(id, StatusCodeVerify, &verifyMsg{
		Flag: authFlag,
		Code: code,
		Msg:  msg,
	})
}

func (m *Manager) emitSignalStatusJSON(id string, code int, msg interface{}) {
	msgStr, err := common.ToJSON(msg)
	if err != nil {
		logger.Warning(err)
		return
	}
	m.emitSignalStatus(id, code, msgStr)
}

type verifyMsg struct {
	Flag int    `json:"flag"`
	Code int    `json:"code"`
	Msg  string `json:"msg"`
}

func (m *Manager) genId() string {
	m.nextIdMu.Lock()
	defer m.nextIdMu.Unlock()

	id := strconv.FormatUint(uint64(m.nextId), 10)
	m.nextId++
	return id
}

func (*Manager) GetInterfaceName() string {
	return dbusServiceIntreface
}

func newManage(service *dbusutil.Service) (*Manager, error) {
	systemConn, err := dbus.SystemBus()
	if err != nil {
		return nil, err
	}

	return &Manager{
		service:        service,
		dbusDaemon:     ofdbus.NewDBus(systemConn),
		sysSigLoop:     dbusutil.NewSignalLoop(systemConn, 10),
		nextId:         1,
		SupportedFlags: AuthenticationFlagPassword | AuthenticationFlagFingerprint | AuthenticationFlagActiveDirectory,
		groups:         make(map[string]*Group),
	}, nil
}

func (m *Manager) init() {
	err := loadConfig(configFile, &m.config)
	if err != nil {
		if !os.IsNotExist(err) {
			logger.Warning(err)
		}
	}
	m.initLimits()
	m.sysSigLoop.Start()
}

func (m *Manager) removeGroup(id string) {
	m.groupsMu.Lock()
	delete(m.groups, id)
	m.groupsMu.Unlock()
}

func (m *Manager) Authenticate(sender dbus.Sender, username string, flag, timeout int32) (result string, busErr *dbus.Error) {
	if username == "" {
		return "", dbusutil.ToError(errors.New("argument username is empty"))
	}
	id := m.genId()
	logger.Debugf("Authenticate sender: %v, username: %q, flag: %v, timeout: %v, id: %v",
		sender, username, flag, timeout, id)

	m.initLimitsByUsername(username)

	cancelLastFingerprintTx()

	ag := newGroup(m, id, username)
	var outFlag int
	if flag&AuthenticationFlagPassword != 0 {
		if m.allowTx(username, AuthTypePassword) {
			tx := newPasswordTx(AuthTypePassword)
			logger.Debugf("auth group %v add PasswordTx password", id)
			ag.addTx(tx)
			outFlag |= AuthenticationFlagPassword
		} else {
			logger.Warning("allow not start password tx")
		}
	}
	if flag&AuthenticationFlagFingerprint != 0 {
		if m.allowTx(username, AuthTypeFingerprint) {
			can, err := hasFingerprintDevice()
			if err != nil {
				logger.Warning(err)
			}
			if can {
				tx := newFingerprintTx()
				err = tx.initTx(username)
				if err == nil {
					outFlag |= AuthenticationFlagFingerprint
					ag.addTx(tx)
					logger.Debugf("auth group %v add FingerprintTx", id)
				} else {
					logger.Warning("failed to init fingerprint tx", err)
				}
			}
		}
	}
	if flag&AuthenticationFlagActiveDirectory != 0 {
		if m.allowTx(username, AuthTypeActiveDirectory) {
			tx := newPasswordTx(AuthTypeActiveDirectory)
			logger.Debugf("auth group %v add PasswordTx active directory", id)
			ag.addTx(tx)
			outFlag |= AuthenticationFlagActiveDirectory
		}
	}

	if outFlag == 0 {
		return "", dbusutil.ToError(errors.New("do not start any tx"))
	}

	if timeout > 0 {
		time.AfterFunc(time.Duration(timeout)*time.Second, func() {
			ag.cancel()
		})
	}
	ag.authenticate()

	m.groupsMu.Lock()
	m.groups[id] = ag
	m.groupsMu.Unlock()

	idFlag := authenticateReturn{
		Id:   id,
		Flag: outFlag,
	}
	result, err := common.ToJSON(&idFlag)
	return result, dbusutil.ToError(err)
}

func (m *Manager) CancelAuthenticate(sender dbus.Sender, id string) (err *dbus.Error) {
	logger.Debugf("CancelAuthenticate sender: %v, id: %v", sender, id)
	m.groupsMu.Lock()
	ag := m.groups[id]
	m.groupsMu.Unlock()
	if ag == nil {
		return dbusutil.ToError(fmt.Errorf("not found auth group %q", id))
	}

	ag.cancel()
	return nil
}

func (m *Manager) PreOneKeyLogin(sender dbus.Sender, flag int32) (string, *dbus.Error) {
	id := m.genId()
	logger.Debugf("Authenticate sender: %v, flag: %v, id: %v",
		sender, flag, id)

	ag := newGroup(m, id, pkgfp.EmptyUsername)
	var outFlag int
	if flag&AuthenticationFlagFingerprint != 0 {
		has, err := hasFingerprintDevice()
		if err != nil {
			logger.Warning(err)
		}
		if has {
			tx := newFingerprintTx()
			err = tx.initTx(pkgfp.EmptyUsername)
			if err == nil {
				ag.addTx(tx)
				outFlag |= AuthenticationFlagFingerprint
				logger.Debugf("auth group %v add FingerprintTx", id)
			} else {
				logger.Warning("failed to init fingerprint tx", err)
			}
		}
	}

	if outFlag == 0 {
		return "", dbusutil.ToError(errors.New("do not start any tx"))
	}

	m.oneKeyLoginWaitCh = make(chan oneKeyLoginResult, 1)

	ag.authenticate()

	m.groupsMu.Lock()
	m.groups[id] = ag
	m.groupsMu.Unlock()

	m.oneKeyLoginRes = oneKeyLoginResult{}
	select {
	case m.oneKeyLoginRes = <-m.oneKeyLoginWaitCh:
		m.oneKeyLoginRes.flag = outFlag
	case <-time.After(1 * time.Second):
	}

	m.oneKeyLoginWaitCh = nil

	ag.cancel()

	return m.oneKeyLoginRes.username, nil
}

func (m *Manager) OneKeyLogin(username string) (string, *dbus.Error) {
	if m.oneKeyLoginRes.username == "" {
		return "", nil
	}

	res := m.oneKeyLoginRes
	m.oneKeyLoginRes = oneKeyLoginResult{}

	sysBus, err := dbus.SystemBus()
	if err != nil {
		logger.Warning("connect to system bus failed:", err)
		return "", dbusutil.ToError(err)
	}

	// 检查当前账户是否有指纹，如果没有，则返回空字符串
	fpObj := authenticate.NewFingerprint(sysBus)
	fingers, err := fpObj.ListFingers(0, username)
	if err != nil {
		return "", dbusutil.ToError(err)
	}
	if len(fingers) == 0 {
		return "", nil
	}

	go func() {
		m.initLimitsByUsername(username)

		// 一键登录验证失败，增加错误次数
		if res.username == pkgfp.EmptyUsername || !res.result {
			logger.Debug("one-key login failed: finger not match")
			m.fail(username, res.txType)
			m.emitOneKeyLoginResult(res.gid, StatusCodeFailure, res.txType)
			return
		}

		limits, err := m.getLimits(username)
		if err != nil {
			logger.Warning("failed to get limits info:", err)
			m.emitOneKeyLoginResult(res.gid, StatusCodeError, res.txType)
			return
		}

		for _, v := range limits {
			if v.Type == res.txType {
				logger.Debugf("one-key login result: %t", !v.Locked)
				code := StatusCodeSuccess
				if v.Locked {
					code = StatusCodeFailure
				}
				m.emitOneKeyLoginResult(res.gid, code, res.txType)
				return
			}
		}

		logger.Debug("one-key login succeeded")
		m.emitOneKeyLoginResult(res.gid, StatusCodeSuccess, res.txType)
	}()

	idFlag := authenticateReturn{
		Id:   res.gid,
		Flag: res.flag,
	}
	result, err := common.ToJSON(&idFlag)
	return result, dbusutil.ToError(err)
}

func (m *Manager) emitOneKeyLoginResult(gid string, code int, txType string) {
	// 延迟发送，避免收不到信号
	time.AfterFunc(500*time.Millisecond, func() {
		m.emitSignalStatus(gid, code, txType)
	})
}

func (m *Manager) SetPassword(sender dbus.Sender, id, password string) (err *dbus.Error) {
	logger.Debugf("SetPassword sender: %v, id: %v, empty: %v", sender, id, password == "")
	m.groupsMu.Lock()
	ag := m.groups[id]
	m.groupsMu.Unlock()
	if ag == nil {
		return dbusutil.ToError(fmt.Errorf("not found auth group %q", id))
	}

	ag.setPassword(password)
	return nil
}

func (m *Manager) GetLimits(username string) (string, *dbus.Error) {
	limits, err := m.getLimits(username)
	if err != nil {
		return "", dbusutil.ToError(err)
	}
	result, err := common.ToJSON(limits)
	logger.Debugf("GetLimits result: %s", result)
	return result, dbusutil.ToError(err)
}

func (m *Manager) setOneKeyLoginResult(gid string, txType string, username string, result bool) {
	if m.oneKeyLoginWaitCh == nil {
		return
	}

	logger.Debug("setOneKeyLoginResult")
	m.oneKeyLoginWaitCh <- oneKeyLoginResult{
		gid:      gid,
		txType:   txType,
		username: username,
		result:   result,
	}
}
