package authenticate

import (
	"errors"
	"fmt"
	"os"

	"pkg.deepin.io/lib/pam"
)

type PasswordTx struct {
	baseTx
	core            *pam.Transaction
	ch              chan string
	setPasswordDone bool
}

func newPasswordTx(type0 string) *PasswordTx {
	tx := &PasswordTx{}
	tx.type0 = type0
	tx.ch = make(chan string, 1)
	return tx
}

func (pt *PasswordTx) setPassword(password string) {
	pt.mu.Lock()
	if !pt.setPasswordDone {
		pt.setPasswordDone = true
		pt.mu.Unlock()
		pt.ch <- password
		return
	}
	pt.mu.Unlock()
}

func (pt *PasswordTx) authenticate() error {
	pamService := ""
	switch pt.type0 {
	case AuthTypePassword:
		pamService = "deepin_pam_unix"
	case AuthTypeActiveDirectory:
		pamService = "deepin_pam_ad"
	default:
		panic(fmt.Errorf("invalid passwordTx type %q", pt.type0))
	}
	var err error
	pt.core, err = pam.Start(pamService, pt.username(), pt)
	if err != nil {
		return err
	}

	go pt.doAuth()
	return nil
}

func (pt *PasswordTx) doAuth() {
	var err error
	if os.Getenv("DEEPIN_AUTH_PASSWORD_TX_FAKE") == "1" {
		err = pt.doAuthFake()
	} else {
		err = pt.doAuthAux()
	}

	var authResult AuthResult = &AuthResultSuccess{}
	if err != nil {
		authResult = &AuthResultFailure{}
	}
	pt.giveResult(pt, authResult)

	err = pt.core.End(pt.core.LastStatus())
	if err != nil {
		logger.Warning(err)
	}
}

func (pt *PasswordTx) doAuthAux() error {
	// meet the requirement of pam_unix.so nullok_secure option,
	// allows any user with a blank password to unlock.
	err := pt.core.SetItemStr(pam.Tty, "tty1")
	if err != nil {
		logger.Warning("failed to set item tty:", err)
	}
	err = pt.core.Authenticate(0)
	return err
}

func (pt *PasswordTx) doAuthFake() error {
	password := <-pt.ch
	switch pt.type0 {
	case AuthTypePassword:
		if password == "pa" {
			return nil
		}
	case AuthTypeActiveDirectory:
		if password == "ad" {
			return nil
		}
	}
	return errors.New("password wrong")
}

func (pt *PasswordTx) RespondPAM(style pam.Style, msg string) (string, error) {
	switch style {
	case pam.PromptEchoOn:
		return pt.username(), nil
	case pam.PromptEchoOff:
		password := <-pt.ch
		return password, nil
	case pam.ErrorMsg:
		logger.Warning("errorMsg:", msg)
		return "", nil
	case pam.TextInfo:
		logger.Warning("textInfo:", msg)
		return "", nil

	default:
		return "", fmt.Errorf("invalid pam style %v", style)
	}
}

func (pt *PasswordTx) end() {
	logger.Debug(pt, "end")
	pt.setPassword("")
}
